//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
// Copyright 2003, 2004, 2005, 2006, 2007, 2008, 2009 James Lehman.
// This source is distributed under the terms of the GNU General Public License.
//
// LaserBoy_frame.cpp is part of LaserBoy.
//
// LaserBoy is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// LaserBoy is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LaserBoy. If not, see <http://www.gnu.org/licenses/>.
//
//############################################################################
#include "LaserBoy_space.hpp"

//############################################################################
LaserBoy_frame::LaserBoy_frame(LaserBoy_space* ps,
                               LaserBoy_vertex(*F)(int, int),
                               int vertices_per_frame,
                               int frame_index
                              )
               : LaserBoy_segment        (ps)
               , transition_to_next_frame(ps)
{
    is_selected            = false;
    is_unique              = true;
    palette_index          = LASERBOY_LASERBOY_HUES;
    error                  = LASERBOY_OK;
    format                 = LASERBOY_3D_FRAME;
    LaserBoy_vertex vertex = F(0, frame_index);
    vertex.blank();
    reserve(vertices_per_frame + 2);
    push_back(vertex);
    for(int p = 0; p < vertices_per_frame; p++)
        push_back(F(p, frame_index));
    push_back(vertex);
    quantity = (u_short)size();
}

//############################################################################
bool LaserBoy_frame::save_as_ild(string file)
{
    ofstream out(file.c_str(), ios::out | ios::binary);
    if(out.is_open())
    {
        if(palette_index != LASERBOY_ILDA_DEFAULT)
          p_space->to_ofstream_ild(out);
        else
            to_ofstream_ild(out);
        out.close();
    }
    else
        return false;
    return true;
}

//############################################################################
bool LaserBoy_frame::save_as_txt(string file)
{
    ofstream out(file.c_str(), ios::out | ios::binary);
    if(out.is_open())
    {
        txt_tag(out);
        if(!p_space->save_txt_with_color)
        {
            if(palette_index != LASERBOY_TRUE_COLOR)
            {
                if(    p_space->save_txt_named_palettes
                    || palette_index >= LASERBOY_BUILT_IN_PALETTES
                  )
                    p_space->palette_picker(palette_index).to_ofstream_txt(out);
                else
                {
                    out << "palette named "
                        << p_space->palette_picker(palette_index).name
                        << endl
                        << endl;
                }
            }
            else
                to_ofstream_txt_table(out);
        }
        to_ofstream_txt(out);
        out.close();
    }
    else
        return false;
    return true;
}

//############################################################################
bool LaserBoy_frame::save_as_txt_table(string file)
{
    ofstream out(file.c_str(), ios::out | ios::binary);
    if(out.is_open())
    {
        txt_tag(out);
        to_ofstream_txt_table(out);
        out.close();
    }
    else
        return false;
    return true;
}

//############################################################################
void LaserBoy_frame::optimize()
{
    minimize                ();
    switch(p_space->dots_setting)
    {
        default:
        case LASERBOY_DOTS_IGNORE:
             break;
        //--------------------------------------------------------------------
        case LASERBOY_DOTS_ENHANCE:
             enhance_dots();
             break;
    }
    add_dwell_vertices      ();
    add_lit_span_vertices   ();
    add_blank_span_vertices ();
    quantity = (u_short)size();
    return;
}

//############################################################################
int LaserBoy_frame::from_ifstream_format_3(ifstream& in, const LaserBoy_ild_header& header)
{
    clear();
    *this = header;
    //------------------------------------------------------------------------
    if(quantity)
    {
        int                  i;
        LaserBoy_color       rgb;
        LaserBoy_vertex      vertex;
        LaserBoy_ild_header  next_header;
        //--------------------------------------------------------------------
        reserve(quantity);
        for(i = 0; i < (int)quantity; i++)
        {
            if(rgb.from_ifstream_ild(in))
                push_back((LaserBoy_vertex)rgb);
            else
                return LASERBOY_BAD_QUANTITY;
        }
        //--------------------------------------------------------------------
        next_header.from_ifstream_ild(in);

        if(quantity != next_header.quantity)
            return LASERBOY_COLOR_VERTEX_MISSMATCH;

        *this = next_header;

        for(i = 0; i < (int)quantity; i++)
        {
            if(vertex.from_ifstream_ild(in, next_header.format))
            {
                at(i).x = vertex.x;
                at(i).y = vertex.y;
                at(i).z = vertex.z;
                at(i).k = vertex.k;
                at(i).c = vertex.c;
            }
            else
                return LASERBOY_BAD_QUANTITY;
        }
        //--------------------------------------------------------------------
        front().blank();
        //--------------------------------------------------------------------
        palette_index = LASERBOY_TRUE_COLOR;
        sync_rgb_and_palette();
        //--------------------------------------------------------------------
    }
    //------------------------------------------------------------------------
    return LASERBOY_OK;
}

//############################################################################
int LaserBoy_frame::from_ifstream_ild(ifstream& in, const LaserBoy_ild_header& header)
{
    int              i;
    LaserBoy_vertex  vertex;
    //------------------------------------------------------------------------
    clear();
    *this = header;
    //------------------------------------------------------------------------
    if(    format == LASERBOY_3D_FRAME_RGB
        || format == LASERBOY_2D_FRAME_RGB
      )
        palette_index = LASERBOY_TRUE_COLOR;
    //------------------------------------------------------------------------
    if(quantity)
    {
        reserve(quantity);
        for(i = 0; i < (int)quantity; i++)
        {
            if(vertex.from_ifstream_ild(in, format))
                push_back(vertex);
//            else
//                return LASERBOY_BAD_QUANTITY;
        }
        quantity = (u_short)size();
        //--------------------------------------------------------------------
        front().blank();
    }
    //------------------------------------------------------------------------
    return LASERBOY_OK;
}

//############################################################################
int LaserBoy_frame::from_ifstream_dxf(ifstream& in)
{
    int              i;
    LaserBoy_vertex  vertex;

    LaserBoy_real_segment real_vertices(p_space);
    //------------------------------------------------------------------------
    format        = LASERBOY_3D_FRAME ;
    palette_index = LASERBOY_TRUE_COLOR;
    //------------------------------------------------------------------------
    real_vertices.from_ifstream_dxf(in);
    //------------------------------------------------------------------------
    if(real_vertices.size() > 2) // more than the default original vector
    {
        real_vertices.normalize();
        reserve(real_vertices.size() - 2);
        for(i = 2; i < (int)real_vertices.size(); i++)
        {
            push_back((LaserBoy_vertex)real_vertices[i]);
            back().set_unique();
        }
    }
    else
    {
        reserve(2);
        push_back(vertex);
        push_back(vertex);
    }
    //------------------------------------------------------------------------
    quantity = (u_short)number_of_vertices();
    is_2D(); // if so change format to LASERBOY_2D_FRAME
    return LASERBOY_OK;
}

//############################################################################
int LaserBoy_frame::from_ifstream_txt(   ifstream& in
                                       , const int& format_1
                                       , const int& format_2
                                       , int& line_number
                                     )
{
    //------------------------------------------------------------------------
    switch(format_1)
    {
        case LASERBOY_TABLE:
                format        = LASERBOY_TABLE;
                palette_index = LASERBOY_TRUE_COLOR;
                break;
        //--------------------------------------------------------------------
        case LASERBOY_3D_FRAME_RGB:
                format        = LASERBOY_3D_FRAME_RGB;
                palette_index = LASERBOY_TRUE_COLOR;
                break;
        //--------------------------------------------------------------------
        case LASERBOY_2D_FRAME_RGB:
                format        = LASERBOY_2D_FRAME_RGB;
                palette_index = LASERBOY_TRUE_COLOR;
                break;
        //--------------------------------------------------------------------
        case LASERBOY_3D_FRAME_HEX:
                format        = LASERBOY_3D_FRAME_RGB;
                palette_index = LASERBOY_TRUE_COLOR;
                break;
        //--------------------------------------------------------------------
        case LASERBOY_2D_FRAME_HEX:
                format        = LASERBOY_2D_FRAME_RGB;
                palette_index = LASERBOY_TRUE_COLOR;
                break;
        //--------------------------------------------------------------------
        case LASERBOY_3D_FRAME_PALETTE:
                format        = LASERBOY_3D_FRAME;
                palette_index = p_space->palette_index;
                break;
        //--------------------------------------------------------------------
        case LASERBOY_2D_FRAME_PALETTE:
                format        = LASERBOY_2D_FRAME;
                palette_index = p_space->palette_index;
                break;
        //--------------------------------------------------------------------
        case LASERBOY_3D_FRAME_TABLE:
                format        = LASERBOY_3D_FRAME;
                palette_index = LASERBOY_TRUE_COLOR;
                break;
        //--------------------------------------------------------------------
        case LASERBOY_2D_FRAME_TABLE:
                format        = LASERBOY_2D_FRAME;
                palette_index = LASERBOY_TRUE_COLOR;
                break;
    }
    //------------------------------------------------------------------------
    if(    format_1 == LASERBOY_3D_FRAME_TABLE
        || format_1 == LASERBOY_2D_FRAME_TABLE // we already have r g b
      )
    {
        LaserBoy_vertex vertex;
        for(int i = 0; i < number_of_vertices(); i++)
            if(!at(i).from_ifstream_txt(in, format_1, format_2, line_number))
                break;
        vertex.from_ifstream_txt(in, format_1, format_2, line_number);
        quantity = (u_short)number_of_vertices();
    }
    //------------------------------------------------------------------------
    else
    {
        LaserBoy_vertex vertex;
        empty(); // clears frame and sets quantity to zero!
        while(vertex.from_ifstream_txt(in, format_1, format_2, line_number))
            push_back(vertex);
        quantity = (u_short)number_of_vertices();
        if(    format_1 == LASERBOY_3D_FRAME_PALETTE
            || format_1 == LASERBOY_2D_FRAME_PALETTE
          )
            set_rgb_from_palette();
    }
    //------------------------------------------------------------------------
    return error;
}

//############################################################################
int LaserBoy_frame::move_selection(LaserBoy_3D_double d)
{
    if(number_of_vertices() > 1)
    {
        LaserBoy_frame   frame(*this);
        LaserBoy_3D_double    f;
        int             i,
                        out_of_bounds = LASERBOY_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                out_of_bounds = LaserBoy_bounds_check(f + d, LASERBOY_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                at(i) = LaserBoy_vertex( f + d,
                                       (LaserBoy_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return LASERBOY_IN_BOUNDS;
}
//############################################################################
int LaserBoy_frame::scale_selection(LaserBoy_3D_double m)
{
    if(number_of_vertices() > 1)
    {
        LaserBoy_3D_double    f,
                        center = selected_segment().find_center();

        int             i,
                        out_of_bounds = LASERBOY_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                out_of_bounds = LaserBoy_bounds_check(((f - center) * m) + center, LASERBOY_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                at(i) = LaserBoy_vertex( ((f - center) * m) + center,
                                       (LaserBoy_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return LASERBOY_IN_BOUNDS;
}

//############################################################################
int LaserBoy_frame::scale_selection_on_coordinates(LaserBoy_3D_double m, LaserBoy_3D_double p)
{
    if(number_of_vertices() > 1)
    {
        LaserBoy_3D_double    f;

        int             i,
                        out_of_bounds = LASERBOY_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                out_of_bounds = LaserBoy_bounds_check(((f - p) * m) + p, LASERBOY_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                at(i) = LaserBoy_vertex( ((f - p) * m) + p,
                                       (LaserBoy_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return LASERBOY_IN_BOUNDS;
}

//############################################################################
int LaserBoy_frame::scale_selection_on_fulcrum(LaserBoy_3D_double magnitude)
{
    if(number_of_vertices() > 1)
    {
        int  i,
             out_of_bounds = LASERBOY_IN_BOUNDS;

        LaserBoy_3D_double  f;
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                out_of_bounds = LaserBoy_bounds_check(((f - p_space->fulcrum) * magnitude) + p_space->fulcrum, LASERBOY_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                at(i) = LaserBoy_vertex( ((f - p_space->fulcrum) * magnitude) + p_space->fulcrum,
                                       (LaserBoy_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return LASERBOY_IN_BOUNDS;
}

//############################################################################
int LaserBoy_frame::rotate_selection(LaserBoy_3D_double a)
{
    if(number_of_vertices() > 1)
    {
        LaserBoy_3D_double    f,
                        center = selected_segment().find_center();

        int             i,
                        out_of_bounds = LASERBOY_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                out_of_bounds = LaserBoy_bounds_check(rotate_vertex_on_coordinates(f, center, a), LASERBOY_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                at(i) = LaserBoy_vertex( rotate_vertex_on_coordinates(f, center, a),
                                       (LaserBoy_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return LASERBOY_IN_BOUNDS;
}

//############################################################################
int LaserBoy_frame::rotate_selection_on_coordinates(LaserBoy_3D_double a, LaserBoy_3D_double p)
{
    if(number_of_vertices() > 1)
    {
        LaserBoy_3D_double    f;

        int             i,
                        out_of_bounds = LASERBOY_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                out_of_bounds = LaserBoy_bounds_check(rotate_vertex_on_coordinates(f, p, a), LASERBOY_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                at(i) = LaserBoy_vertex( rotate_vertex_on_coordinates(f, p, a),
                                       (LaserBoy_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return LASERBOY_IN_BOUNDS;
}

//############################################################################
int LaserBoy_frame::rotate_selection_on_fulcrum(LaserBoy_3D_double angle)
{
    if(number_of_vertices() > 1)
    {
        LaserBoy_3D_double    f;

        int             i,
                        out_of_bounds = LASERBOY_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                out_of_bounds = LaserBoy_bounds_check(rotate_vertex_on_coordinates(f, p_space->fulcrum, angle), LASERBOY_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_index_selected(i))
            {
                f = at(i).position();
                at(i) = LaserBoy_vertex( rotate_vertex_on_coordinates(f, p_space->fulcrum, angle),
                                       (LaserBoy_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return LASERBOY_IN_BOUNDS;
}

//############################################################################
void LaserBoy_frame::next_segment_select()
{
    int i, start, end;
    find_segment(egg, start, end, i);
    find_segment(i + 1, egg, spider);
    return;
}

//############################################################################
void LaserBoy_frame::next_segment_egg()
{
    int i, start, end;
    find_segment(egg, start, end, i);
    find_segment(i + 1, egg, end);
    return;
}

//############################################################################
void LaserBoy_frame::next_segment_spider()
{
    int i, start, end;
    find_segment(spider, start, end, i);
    find_segment(i + 1, start, spider);
    return;
}

//############################################################################
void LaserBoy_frame::previous_segment_select()
{
    int i, start, end;
    find_segment(egg, start, end, i);
    find_segment(i - 1, egg, spider);
    return;
}

//############################################################################
void LaserBoy_frame::previous_segment_egg()
{
    int i, start, end;
    find_segment(egg, start, end, i);
    find_segment(i - 1, egg, end);
    return;
}

//############################################################################
void LaserBoy_frame::previous_segment_spider()
{
    int i, start, end;
    find_segment(spider, start, end, i);
    find_segment(i - 1, start, spider);
    return;
}

//############################################################################
void LaserBoy_frame::color_select()
{
    if(quantity > 1)
    {
        int i;
        //--------------------------------------------------------------------
        if(palette_index != LASERBOY_TRUE_COLOR)
        {
            for(i = 0; i < number_of_vertices(); i++)
                if(    at(i).is_color(p_space->black_level)
                    && is_index_selected(i)
                  )
                {
                    at(i).c = p_space->selected_color_index;
                    at(i).r = p_space->palette_picker(palette_index)[p_space->selected_color_index].r;
                    at(i).g = p_space->palette_picker(palette_index)[p_space->selected_color_index].g;
                    at(i).b = p_space->palette_picker(palette_index)[p_space->selected_color_index].b;
                }
        }
        //--------------------------------------------------------------------
        else
        {
            for(i = 0; i < number_of_vertices(); i++)
                if(    at(i).is_color(p_space->black_level)
                    && is_index_selected(i)
                  )
                {
                    at(i).r = p_space->palette_picker(palette_index)[p_space->selected_color_index].r;
                    at(i).g = p_space->palette_picker(palette_index)[p_space->selected_color_index].g;
                    at(i).b = p_space->palette_picker(palette_index)[p_space->selected_color_index].b;
                }
        }
        //--------------------------------------------------------------------
    }
    return;
}

//############################################################################
void LaserBoy_frame::color_index_select()
{
    if(quantity > 1)
    {
        int i,
            j = 0;
        //--------------------------------------------------------------------
        if(palette_index != LASERBOY_TRUE_COLOR)
        {
            for(i = 0; i < number_of_vertices(); i++)
                if(    at(i).is_color(p_space->black_level)
                    && is_index_selected(i)
                  )
                {
                    at(i).c = p_space->selected_color_index;
                    at(i).r = p_space->palette_picker(palette_index)[j % p_space->palette_picker(palette_index).last].r;
                    at(i).g = p_space->palette_picker(palette_index)[j % p_space->palette_picker(palette_index).last].g;
                    at(i).b = p_space->palette_picker(palette_index)[j % p_space->palette_picker(palette_index).last].b;
                    j++;
                }
        }
        //--------------------------------------------------------------------
        else
        {
            for(i = 0; i < number_of_vertices(); i++)
                if(    at(i).is_color(p_space->black_level)
                    && is_index_selected(i)
                  )
                {
                    at(i).r = p_space->palette_picker(p_space->target_palette_index)[j % p_space->palette_picker(p_space->target_palette_index).last].r;
                    at(i).g = p_space->palette_picker(p_space->target_palette_index)[j % p_space->palette_picker(p_space->target_palette_index).last].g;
                    at(i).b = p_space->palette_picker(p_space->target_palette_index)[j % p_space->palette_picker(p_space->target_palette_index).last].b;
                    j++;
                }
        }
        //--------------------------------------------------------------------
    }
    return;
}

//############################################################################
void LaserBoy_frame::color_span_select()
{
    if(quantity > 1)
    {
        int i,
            j = 0,
            selected = 0,
            span = egg <= spider ? egg : spider,
            offset = egg <= spider ? spider : egg;
        //--------------------------------------------------------------------
        for(i = span; i <= offset; i++)
            if(at(i).is_color(p_space->black_level))
                selected++;
        //--------------------------------------------------------------------
        if(palette_index != LASERBOY_TRUE_COLOR)
        {
            span   =    p_space->palette_picker(palette_index).last
                      - p_space->palette_picker(palette_index).first,
            offset =    p_space->palette_picker(palette_index).first;
            for(i = 0; i < number_of_vertices(); i++)
                if(    at(i).is_color(p_space->black_level)
                    && is_index_selected(i)
                  )
                {
                    at(i).c = (u_char)  (   j
                                          * p_space->recolor_span_factor
                                          / selected
                                          * span
                                        )
                                      % span
                                      + offset;
                    at(i).r = p_space->palette_picker(palette_index)[at(i).c].r;
                    at(i).g = p_space->palette_picker(palette_index)[at(i).c].g;
                    at(i).b = p_space->palette_picker(palette_index)[at(i).c].b;
                    j++;
                }
        }
        //--------------------------------------------------------------------
        else
        {
            span   =    p_space->palette_picker(p_space->target_palette_index).last
                      - p_space->palette_picker(p_space->target_palette_index).first,
            offset =    p_space->palette_picker(p_space->target_palette_index).first;
            for(i = 0; i < number_of_vertices(); i++)
                if(    at(i).is_color(p_space->black_level)
                    && is_index_selected(i)
                  )
                {
                    at(i).c = (u_char)  (   j
                                          * p_space->recolor_span_factor
                                          / selected
                                          * span
                                        )
                                      % span
                                      + offset;
                    at(i).r = p_space->palette_picker(p_space->target_palette_index)[at(i).c].r;
                    at(i).g = p_space->palette_picker(p_space->target_palette_index)[at(i).c].g;
                    at(i).b = p_space->palette_picker(p_space->target_palette_index)[at(i).c].b;
                    j++;
                }
        }
        //--------------------------------------------------------------------
    }
    return;
}

//############################################################################
void LaserBoy_frame::color_index_black()
{
    if(quantity > 1)
    {
        int i,
            j = 0;
        //--------------------------------------------------------------------
        if(palette_index != LASERBOY_TRUE_COLOR)
        {
            for(i = 0; i < number_of_vertices(); i++)
                if(    at(i).is_black(p_space->black_level)
                    && is_index_selected(i)
                  )
                {
                    at(i).c = p_space->selected_color_index;
                    at(i).r = p_space->palette_picker(palette_index)[j % p_space->palette_picker(palette_index).last].r;
                    at(i).g = p_space->palette_picker(palette_index)[j % p_space->palette_picker(palette_index).last].g;
                    at(i).b = p_space->palette_picker(palette_index)[j % p_space->palette_picker(palette_index).last].b;
                    j++;
                }
        }
        //--------------------------------------------------------------------
        else
        {
            for(i = 0; i < number_of_vertices(); i++)
                if(    at(i).is_black(p_space->black_level)
                    && is_index_selected(i)
                  )
                {
                    at(i).r = p_space->palette_picker(p_space->target_palette_index)[j % p_space->palette_picker(p_space->target_palette_index).last].r;
                    at(i).g = p_space->palette_picker(p_space->target_palette_index)[j % p_space->palette_picker(p_space->target_palette_index).last].g;
                    at(i).b = p_space->palette_picker(p_space->target_palette_index)[j % p_space->palette_picker(p_space->target_palette_index).last].b;
                    j++;
                }
        }
        //--------------------------------------------------------------------
    }
    return;
}

//############################################################################
void LaserBoy_frame::color_span_black()
{
    if(quantity > 1)
    {
        int i,
            j = 0,
            selected = 0,
            span = egg <= spider ? egg : spider,
            offset = egg <= spider ? spider : egg;
        //--------------------------------------------------------------------
        for(i = span; i <= offset; i++)
            if(at(i).is_black(p_space->black_level))
                selected++;
        //--------------------------------------------------------------------
        if(palette_index != LASERBOY_TRUE_COLOR)
        {
            span   =    p_space->palette_picker(palette_index).last
                      - p_space->palette_picker(palette_index).first,
            offset =    p_space->palette_picker(palette_index).first;
            for(i = 0; i < number_of_vertices(); i++)
                if(    at(i).is_black(p_space->black_level)
                    && is_index_selected(i)
                  )
                {
                    at(i).c = (u_char)  (   j
                                          * p_space->recolor_span_factor
                                          / selected
                                          * span
                                        )
                                      % span
                                      + offset;
                    at(i).r = p_space->palette_picker(palette_index)[at(i).c].r;
                    at(i).g = p_space->palette_picker(palette_index)[at(i).c].g;
                    at(i).b = p_space->palette_picker(palette_index)[at(i).c].b;
                    j++;
                }
        }
        //--------------------------------------------------------------------
        else
        {
            span   =    p_space->palette_picker(p_space->target_palette_index).last
                      - p_space->palette_picker(p_space->target_palette_index).first,
            offset =    p_space->palette_picker(p_space->target_palette_index).first;
            for(i = 0; i < number_of_vertices(); i++)
                if(    at(i).is_black(p_space->black_level)
                    && is_index_selected(i)
                  )
                {
                    at(i).c = (u_char)  (   j
                                          * p_space->recolor_span_factor
                                          / selected
                                          * span
                                        )
                                      % span
                                      + offset;
                    at(i).r = p_space->palette_picker(p_space->target_palette_index)[at(i).c].r;
                    at(i).g = p_space->palette_picker(p_space->target_palette_index)[at(i).c].g;
                    at(i).b = p_space->palette_picker(p_space->target_palette_index)[at(i).c].b;
                    j++;
                }
        }
        //--------------------------------------------------------------------
    }
    return;
}

//############################################################################
void LaserBoy_frame::color_black_in_select()
{
    if(quantity > 1)
    {
        int i;
        //--------------------------------------------------------------------
        if(palette_index != LASERBOY_TRUE_COLOR)
        {
            for(i = 0; i < number_of_vertices(); i++)
                if(is_index_selected(i) && at(i).is_black(p_space->black_level))
                {
                    at(i).c = p_space->selected_color_index;
                    at(i).r = p_space->palette_picker(palette_index)[p_space->selected_color_index].r;
                    at(i).g = p_space->palette_picker(palette_index)[p_space->selected_color_index].g;
                    at(i).b = p_space->palette_picker(palette_index)[p_space->selected_color_index].b;
                }
        }
        //--------------------------------------------------------------------
        else
        {
            for(i = 0; i < number_of_vertices(); i++)
                if(is_index_selected(i) && at(i).is_black(p_space->black_level))
                {
                    at(i).r = p_space->palette_picker(palette_index)[p_space->selected_color_index].r;
                    at(i).g = p_space->palette_picker(palette_index)[p_space->selected_color_index].g;
                    at(i).b = p_space->palette_picker(palette_index)[p_space->selected_color_index].b;
                }
        }
        //--------------------------------------------------------------------
    }
    return;
}

//############################################################################
void LaserBoy_frame::color_cycle_select(int steps)
{
    if(quantity > 1)
    {
        int i,
            span,
            offset;
        //--------------------------------------------------------------------
        if(palette_index != LASERBOY_TRUE_COLOR)
        {
            span   =    p_space->palette_picker(palette_index).last
                      - p_space->palette_picker(palette_index).first,
            offset =    p_space->palette_picker(palette_index).first;
            while(span > 0 && steps < 0)
                steps += span;
            for(i = 0; i < number_of_vertices(); i++)
                if(    at(i).is_color(p_space->black_level)
                    && is_index_selected(i)
                  )
                {
                    at(i).c = (at(i).c - offset + steps) % span + offset;
                    at(i).r = p_space->palette_picker(palette_index)[at(i).c].r;
                    at(i).g = p_space->palette_picker(palette_index)[at(i).c].g;
                    at(i).b = p_space->palette_picker(palette_index)[at(i).c].b;
                }
        }
        //--------------------------------------------------------------------
        else
        {
            LaserBoy_segment selected_vectors;
            vector<int>      selected_vectors_index;
            
            for(i = 0; i < number_of_vertices(); i++)
                if(    at(i).is_color(p_space->black_level)
                    && is_index_selected(i)
                  )
                {
                    selected_vectors.push_back(at(i));
                    selected_vectors_index.push_back(i);
                }
                
            while(selected_vectors.size() > 0 && steps < 0)
                steps += selected_vectors.size();
                
            for(i = 0; i < selected_vectors.size(); i++)
            {
                at(selected_vectors_index[i]).r = selected_vectors[(i + steps) % selected_vectors.size()].r;
                at(selected_vectors_index[i]).g = selected_vectors[(i + steps) % selected_vectors.size()].g;
                at(selected_vectors_index[i]).b = selected_vectors[(i + steps) % selected_vectors.size()].b;
            }
        }
        //--------------------------------------------------------------------
    }
    return;
}

//############################################################################
void LaserBoy_frame::blank_vertices()
{
    if(quantity > 1)
    {
        for(int i = 0; i < number_of_vertices(); i++)
            if(is_index_selected(i))
                at(i).blank();
    }
    return;
}

//############################################################################
void LaserBoy_frame::unblank_vertices()
{
    if(quantity > 1)
    {
        for(int i = 1; i < number_of_vertices(); i++)
            if(is_index_selected(i))
                at(i).unblank();
    }
    return;
}

//############################################################################
void LaserBoy_frame::black_vertices()
{
    if(quantity > 1)
    {
        for(int i = 1; i < number_of_vertices(); i++)
            if(is_index_selected(i) && at(i).is_blank())
            {
                at(i).r = at(i).g = at(i).b = 0x00;
                at(i).c = p_space->palette_picker(palette_index).black;
                at(i).unblank();
            }
    }
    return;
}

//############################################################################
void LaserBoy_frame::unblack_vertices()
{
    if(quantity > 1)
    {
        for(int i = 1; i < number_of_vertices(); i++)
            if(is_index_selected(i) && at(i).color() == 0x00) // blank
                at(i).blank();
    }
    return;
}

//############################################################################
void LaserBoy_frame::blank_equivalent_vectors()
{
    LaserBoy_frame_set  segments(p_space);
    int                 i,
                        j;
    //------------------------------------------------------------------------
    fracture_segments();
    segments = find_segments();
    //------------------------------------------------------------------------
    for(i = 0; i < segments.number_of_frames() - 1; i++)
    {
        for(j = i + 1; j < segments.number_of_frames(); j++)
            if(    segments[j] == segments[i]
                || segments[j] == segments[i].reverse()
              )
                segments[j].blank_all_vertices();
    }
    //------------------------------------------------------------------------
    *this = segments.sum_of_frames();
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void LaserBoy_frame::minimize()
{
    if(!p_space->allow_lit_black)
        convert_black_to_blank();
    //------------------------------------------------------------------------
    remove_dwell_vertices();
    reduce_blank_vectors ();
    reduce_lit_vectors   ();
    //------------------------------------------------------------------------
    if(p_space->no_equivalent_vectors)
    {
        blank_equivalent_vectors();
        reduce_blank_vectors    ();
    }
    //------------------------------------------------------------------------
    switch(p_space->dots_setting)
    {
        default:
        case LASERBOY_DOTS_IGNORE:
             break;
        //--------------------------------------------------------------------
        case LASERBOY_DOTS_REMOVE:
             remove_dots();
             break;
    }
    //------------------------------------------------------------------------
    bond_segments();
    validate     ();
    //------------------------------------------------------------------------
    return;
}

//############################################################################
LaserBoy_frame& LaserBoy_frame::add_new_vertex(int vertex_index)
{
    LaserBoy_segment::add_new_vertex(vertex_index);
    quantity = (u_short)size();
    if(    quantity > 2
        && vertex_index >= 0
        && vertex_index <  quantity
      )
    {
        if((at(vertex_index) | at(vertex_index + 1)) > p_space->insignificant_distance)
            spider = vertex_index + 1;
        else
            spider = vertex_index;
    }
    else
        egg = spider = vertex_index = 0;
    return *this;
}

//############################################################################
LaserBoy_frame& LaserBoy_frame::break_segment()
{
    LaserBoy_segment::break_segment(spider);
    quantity = (u_short)size();
    return *this;
}

//############################################################################
LaserBoy_frame& LaserBoy_frame::set_spider_to_egg()
{
    at(spider) = (LaserBoy_3D_short)at(egg);
    return *this;
}

//############################################################################
LaserBoy_frame& LaserBoy_frame::set_egg_to_spider()
{
    at(egg) = (LaserBoy_3D_short)at(spider);
    return *this;
}

//############################################################################
LaserBoy_frame& LaserBoy_frame::connect_the_dots()
{
    LaserBoy_segment::connect_the_dots(egg, spider);
    quantity = (u_short)size();
    return *this;
}

//############################################################################
LaserBoy_frame& LaserBoy_frame::to_dots()
{
    int             i;
    LaserBoy_frame  frame(p_space, palette_index);
    //------------------------------------------------------------------------
    frame.reserve(number_of_vertices() + size_of_selection());
    //------------------------------------------------------------------------
    for(i = 0; i < number_of_vertices(); i++)
    {
        if(    at(i).is_lit()
            && at_index_has_magnitude(i)
          )
        {
            frame += at(i);
            frame.back().r = 0;
            frame.back().g = 0;
            frame.back().b = 0;
            frame.back().c = p_space->palette_picker(palette_index).black;
        }
        frame += at(i);
    }
    //------------------------------------------------------------------------
    *this = frame;
    quantity = (u_short)size();
    //------------------------------------------------------------------------
    return *this;
}

//############################################################################
LaserBoy_frame& LaserBoy_frame::selected_to_dots()
{
    int             i,
                    new_spider = spider;
    LaserBoy_frame  frame(p_space, palette_index);
    //------------------------------------------------------------------------
    frame.reserve(number_of_vertices() + size_of_selection());
    //------------------------------------------------------------------------
    frame.push_back(front());
    for(i = 1; i < number_of_vertices(); i++)
    {
        if(    is_index_selected(i)
            && at(i).is_lit()
            && at_index_has_magnitude(i)
          )
        {
            frame += at(i);
            frame.back().r = 0;
            frame.back().g = 0;
            frame.back().b = 0;
            frame.back().c = p_space->palette_picker(palette_index).black;
            new_spider++;
        }
        frame += at(i);
    }
    //------------------------------------------------------------------------
    spider = new_spider;
    *this = frame;
    quantity = (u_short)size();
    //------------------------------------------------------------------------
    return *this;
}


//############################################################################
LaserBoy_frame& LaserBoy_frame::remove_vertex(int vertex_index)
{
    LaserBoy_segment::remove_vertex(vertex_index);
    validate();
    return *this;
}

//############################################################################
void LaserBoy_frame::to_ofstream_ild(ofstream& out)
{
    char number[7];
    if(p_space->dumb_file_format)
    {
        if(format == LASERBOY_3D_FRAME)
            format = LASERBOY_3D_FRAME_RGB;
        else if(format == LASERBOY_2D_FRAME)
            format = LASERBOY_2D_FRAME_RGB;
        sprintf(number, "%06d", identity);
        name = "ES" + string(number);
        LaserBoy_ild_header::to_ofstream_ild(out);
        for(int i = 0; i < (int)quantity; i++)
            at(i).to_ofstream_ild(out, format);
    }
    else
    {
        if(palette_index == LASERBOY_TRUE_COLOR)
        {
            LaserBoy_ild_header header  = (LaserBoy_ild_header)*this;
            LaserBoy_palette    palette = as_color_table();
            header.identity--;
            sprintf(number, "%06d", header.identity);
            header.format    = LASERBOY_TABLE;
            header.name      = "ES" + string(number);
            palette.to_ofstream_ild(out, header);
        }
        else
        {
            sprintf(number, "%06d", identity);
            name = "ES" + string(number);
        }
        LaserBoy_ild_header::to_ofstream_ild(out);
        for(int i = 0; i < (int)quantity; i++)
            at(i).to_ofstream_ild(out, format);
    }
    return;
}

//############################################################################
void LaserBoy_frame::to_ofstream_txt_table(ofstream& out)
{
    out << "table ";
    if(p_space->save_txt_color_hex)
        out << "hex";
    else
        out << "rgb";
    out << endl;
    //------------------------------------------------------------------------
    for(int i = 0; i < quantity; i++)
        if(at(i).is_lit())
            ((LaserBoy_color)at(i)).to_ofstream_txt(out, p_space->save_txt_color_hex);
        else
            out << "    -1"
                << endl;
    //------------------------------------------------------------------------
    out << endl;
    return;
}

//############################################################################
void LaserBoy_frame::to_ofstream_txt(ofstream& out)
{
    int output_format = (   (is_2D())
                           ?(  (p_space->save_txt_with_color)
                              ?(LASERBOY_2D_FRAME_RGB)
                              :(   (palette_index == LASERBOY_TRUE_COLOR)
                                 ? (LASERBOY_2D_FRAME_TABLE)
                                 : (LASERBOY_2D_FRAME_PALETTE)
                               )
                            ) // 3D
                           :(  (p_space->save_txt_with_color)
                              ?(LASERBOY_3D_FRAME_RGB)
                              :(   (palette_index == LASERBOY_TRUE_COLOR)
                                 ? (LASERBOY_3D_FRAME_TABLE)
                                 : (LASERBOY_3D_FRAME_PALETTE)
                               )
                            )
                        );

    out << "frame ";
    //------------------------------------------------------------------------
    switch(output_format)
    {
        case LASERBOY_2D_FRAME_RGB:
        case LASERBOY_2D_FRAME_TABLE:
        case LASERBOY_2D_FRAME_PALETTE:
                out << "xy ";
                break;
        case LASERBOY_3D_FRAME_RGB:
        case LASERBOY_3D_FRAME_TABLE:
        case LASERBOY_3D_FRAME_PALETTE:
                out << "xyz ";
                break;
    }
    //------------------------------------------------------------------------
    switch(output_format)
    {
        case LASERBOY_2D_FRAME_RGB:
        case LASERBOY_3D_FRAME_RGB:
                if(p_space->save_txt_color_hex)
                    out << "hex";
                else
                    out << "rgb";
                break;
        case LASERBOY_2D_FRAME_TABLE:
        case LASERBOY_3D_FRAME_TABLE:
                out << "table";
                break;
        case LASERBOY_2D_FRAME_PALETTE:
        case LASERBOY_3D_FRAME_PALETTE:
                out << "palette";
                break;
    }
    //------------------------------------------------------------------------
    if(p_space->save_txt_unit)
    {
        out << " unit" << endl;
        switch(output_format)
        {
            case LASERBOY_3D_FRAME_PALETTE:
                     for(int i = 0; i < (int)quantity; i++)
                         out << "    "
                             << showpos
                             << setw(15) << left << at(i).x / (double)LASERBOY_MAX_SHORT
                             << setw(15) << left << at(i).y / (double)LASERBOY_MAX_SHORT
                             << setw(15) << left << at(i).z / (double)LASERBOY_MAX_SHORT
                             << noshowpos
                             << setw(7) << ((at(i).is_lit()) ? ((int)at(i).c) : (-1))
                             << endl;
                     break;
            //----------------------------------------------------------------
            case LASERBOY_2D_FRAME_PALETTE:
                     for(int i = 0; i < (int)quantity; i++)
                         out << "    "
                             << showpos
                             << setw(15) << left << at(i).x / (double)LASERBOY_MAX_SHORT
                             << setw(15) << left << at(i).y / (double)LASERBOY_MAX_SHORT
                             << noshowpos
                             << setw(7) << ((at(i).is_lit()) ? ((int)at(i).c) : (-1))
                             << endl;
                     break;
            //----------------------------------------------------------------
            case LASERBOY_3D_FRAME_TABLE: // nothing but x y z
                     for(int i = 0; i < (int)quantity; i++)
                         out << "    "
                             << showpos
                             << setw(15) << left << at(i).x / (double)LASERBOY_MAX_SHORT
                             << setw(15) << left << at(i).y / (double)LASERBOY_MAX_SHORT
                             << setw(15) << left << at(i).z / (double)LASERBOY_MAX_SHORT
                             << noshowpos
                             << endl;
                     break;
            //----------------------------------------------------------------
            case LASERBOY_2D_FRAME_TABLE: // nothing but x y
                     for(int i = 0; i < (int)quantity; i++)
                         out << "    "
                             << showpos
                             << setw(15) << left << at(i).x / (double)LASERBOY_MAX_SHORT
                             << setw(15) << left << at(i).y / (double)LASERBOY_MAX_SHORT
                             << noshowpos
                             << endl;
                     break;
            //----------------------------------------------------------------
            case LASERBOY_3D_FRAME_RGB:
                     for(int i = 0; i < (int)quantity; i++)
                     {
                         out << "    "
                             << showpos
                             << setw(15) << left << at(i).x / (double)LASERBOY_MAX_SHORT
                             << setw(15) << left << at(i).y / (double)LASERBOY_MAX_SHORT
                             << setw(15) << left << at(i).z / (double)LASERBOY_MAX_SHORT
                             << noshowpos;
                         if(p_space->save_txt_color_hex)
                         {
                             if(at(i).is_lit())
                                 out << "    0x"
                                     << hex
                                     << setw(2) << setfill('0') << (int)at(i).r
                                     << setw(2) << setfill('0') << (int)at(i).g
                                     << setw(2) << setfill('0') << (int)at(i).b
                                     << setfill(' ');
                             else
                                 out << dec
                                     << "   -1";
                         }
                         else
                         {
                             if(at(i).is_lit())
                                 out << setw(7) << (int)at(i).r
                                     << setw(7) << (int)at(i).g
                                     << setw(7) << (int)at(i).b;
                             else
                                 out << dec
                                     << "   -1";
                         }
                         out << endl;
                     }
                     break;
            //----------------------------------------------------------------
            case LASERBOY_2D_FRAME_RGB:
                     for(int i = 0; i < (int)quantity; i++)
                     {
                         out << "    "
                             << showpos
                             << setw(15) << left << at(i).x / (double)LASERBOY_MAX_SHORT
                             << setw(15) << left << at(i).y / (double)LASERBOY_MAX_SHORT
                             << noshowpos;
                         if(p_space->save_txt_color_hex)
                         {
                             if(at(i).is_lit())
                                 out << "    0x"
                                     << hex
                                     << setw(2) << setfill('0') << (int)at(i).r
                                     << setw(2) << setfill('0') << (int)at(i).g
                                     << setw(2) << setfill('0') << (int)at(i).b
                                     << setfill(' ');
                             else
                                 out << dec
                                     << "   -1";
                         }
                         else
                         {
                             if(at(i).is_lit())
                                 out << setw(7) << (int)at(i).r
                                     << setw(7) << (int)at(i).g
                                     << setw(7) << (int)at(i).b;
                             else
                                 out << dec
                                     << "   -1";
                         }
                         out << endl;
                     }
                     break;
            //----------------------------------------------------------------
        }
    }
    else
    {
        out << " short" << endl;
        switch(output_format)
        {
            case LASERBOY_3D_FRAME_PALETTE:
                     for(int i = 0; i < (int)quantity; i++)
                         out << setw(9) << at(i).x
                             << setw(9) << at(i).y
                             << setw(9) << at(i).z
                             << setw(7) << ((at(i).is_lit()) ? ((int)at(i).c) : (-1))
                             << endl;
                     break;
            //----------------------------------------------------------------
            case LASERBOY_2D_FRAME_PALETTE:
                     for(int i = 0; i < (int)quantity; i++)
                         out << setw(9) << at(i).x
                             << setw(9) << at(i).y
                             << setw(7) << ((at(i).is_lit()) ? ((int)at(i).c) : (-1))
                             << endl;
                     break;
            //----------------------------------------------------------------
            case LASERBOY_3D_FRAME_TABLE:
                     for(int i = 0; i < (int)quantity; i++)
                         out << setw(9) << at(i).x
                             << setw(9) << at(i).y
                             << setw(9) << at(i).z
                             << endl;
                     break;
            //----------------------------------------------------------------
            case LASERBOY_2D_FRAME_TABLE:
                     for(int i = 0; i < (int)quantity; i++)
                         out << setw(9) << at(i).x
                             << setw(9) << at(i).y
                             << endl;
                     break;
            //----------------------------------------------------------------
            case LASERBOY_3D_FRAME_RGB:
                     for(int i = 0; i < (int)quantity; i++)
                     {
                         out << dec
                             << setw(9) << at(i).x
                             << setw(9) << at(i).y
                             << setw(9) << at(i).z;
                         if(p_space->save_txt_color_hex)
                         {
                             if(at(i).is_lit())
                                 out << "    0x"
                                     << hex
                                     << setw(2) << setfill('0') << (int)at(i).r
                                     << setw(2) << setfill('0') << (int)at(i).g
                                     << setw(2) << setfill('0') << (int)at(i).b
                                     << setfill(' ');
                             else
                                 out << dec
                                     << "   -1";
                         }
                         else
                         {
                             if(at(i).is_lit())
                                 out << setw(7) << (int)at(i).r
                                     << setw(7) << (int)at(i).g
                                     << setw(7) << (int)at(i).b;
                             else
                                 out << dec
                                     << "   -1";
                         }
                         out << endl;
                     }
                     break;
            //----------------------------------------------------------------
            case LASERBOY_2D_FRAME_RGB:
                     for(int i = 0; i < (int)quantity; i++)
                     {
                         out << dec
                             << setw(9) << at(i).x
                             << setw(9) << at(i).y;
                         if(p_space->save_txt_color_hex)
                         {
                             if(at(i).is_lit())
                                 out << "    0x"
                                     << hex
                                     << setw(2) << setfill('0') << (int)at(i).r
                                     << setw(2) << setfill('0') << (int)at(i).g
                                     << setw(2) << setfill('0') << (int)at(i).b
                                     << setfill(' ');
                             else
                                 out << dec
                                     << "   -1";
                         }
                         else
                         {
                             if(at(i).is_lit())
                                 out << setw(7) << (int)at(i).r
                                     << setw(7) << (int)at(i).g
                                     << setw(7) << (int)at(i).b;
                             else
                                 out << dec
                                     << "   -1";
                         }
                         out << endl;
                     }
                     break;
            //----------------------------------------------------------------
        }
    }
    out << endl;
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void LaserBoy_frame::render(int skin) const
{
    bool     rotate_view  = (p_space->view_angle  != 0.0),
             offset_view  = (p_space->view_offset != 0.0),
             scale_view   = (p_space->view_scale  != 1.0);

    u_short  i;

    int      xx,
             yy,
             x_off        = p_space->bmp.yres + LASERBOY_STATS_DISPLAY_OFFSET * p_space->font_size_factor,
             y_off        =   LASERBOY_PALETTE_MAX * p_space->font_size_factor
                            + (LASERBOY_STATS_DISPLAY_OFFSET + 24) * p_space->font_size_factor,
             y_line       = 8 * p_space->font_size_factor,
             bmp_center   = p_space->bmp.yres / 2;

    u_int    bmp_bg_color = p_space->bmp.make_rgb(p_space->bg_color.r, p_space->bg_color.g, p_space->bg_color.b);

    double   scale        = p_space->bmp.yres / 65536.0,
             fade;

    LaserBoy_3D_double _0,
                       _1;

    LaserBoy_color     color;
    //------------------------------------------------------------------------
    // ACTUAL VECTOR ART !!!
    //------------------------------------------------------------------------
    if(quantity > 1)
    {
        if(    p_space->show_blanking
            && (    skin == 0
                 || p_space->show_onion_visuals
               )
          )
        {
            for(i = 1; i < quantity; i++)
            {
                if(at(i).is_blank())
                {
                    if(rotate_view)
                    {
                        _0 = rotate_vertex(at(i - 1).bit_masked_position(p_space->signal_bit_mask), p_space->view_angle);
                        _1 = rotate_vertex(at(i    ).bit_masked_position(p_space->signal_bit_mask), p_space->view_angle);
                    }
                    else
                    {
                        _0 = at(i - 1).bit_masked_position(p_space->signal_bit_mask);
                        _1 = at(i    ).bit_masked_position(p_space->signal_bit_mask);
                    }
                    if(offset_view)
                    {
                        _0 += p_space->view_offset;
                        _1 += p_space->view_offset;
                    }
                    if(scale_view)
                    {
                        _0 *= p_space->view_scale;
                        _1 *= p_space->view_scale;
                    }
                    xx = (int)(_1.x * scale) + bmp_center;
                    yy = (int)(_1.y * scale) + bmp_center;
                    
                    color = p_space->bmp.make_rgb(128, 128, 128);
                    if(skin)
                    {
                        fade = p_space->onion_skin_fade + (skin - 1) * (p_space->onion_skin_fade / 10);
                        if(fade > 1.0)
                            fade = 1.0;
                        color = color.blend(bmp_bg_color, fade);
                    }
                    bmp_put_line_fast(   &(p_space->bmp),     (int)(_0.x * scale) + bmp_center
                                       , p_space->bmp.yres - ((int)(_0.y * scale) + bmp_center)
                                       , xx
                                       , p_space->bmp.yres - (yy)
                                       , color.as_uint()
                                     );
                    if(    p_space->show_vertices
                        && (    skin == 0
                             || p_space->show_onion_visuals
                           )
                      )
                        bmp_put_vertex(   &(p_space->bmp)
                                        , xx
                                        , p_space->bmp.yres - yy
                                        , color.as_uint()
                                      );
                    if(    p_space->show_cursors
                        && is_index_selected(i)
                        && (    skin == 0
                             || p_space->show_onion_visuals
                           )
                      )
                        bmp_put_select(&(p_space->bmp)
                                        , xx
                                        , p_space->bmp.yres - yy
                                        , color.as_uint()
                                      );
                }
            }
        } // end if(p_space->show_blanking)
        //--------------------------------------------------------------------
        for(i = 1; i < quantity; i++)
        {
            if(at(i).is_black(p_space->black_level))
            {
                if(rotate_view)
                {
                    _0 = rotate_vertex(at(i - 1).bit_masked_position(p_space->signal_bit_mask), p_space->view_angle);
                    _1 = rotate_vertex(at(i    ).bit_masked_position(p_space->signal_bit_mask), p_space->view_angle);
                }
                else
                {
                    _0 = at(i - 1).bit_masked_position(p_space->signal_bit_mask);
                    _1 = at(i    ).bit_masked_position(p_space->signal_bit_mask);
                }
                if(offset_view)
                {
                    _0 += p_space->view_offset;
                    _1 += p_space->view_offset;
                }
                if(scale_view)
                {
                    _0 *= p_space->view_scale;
                    _1 *= p_space->view_scale;
                }
                xx = (int)(_1.x * scale) + bmp_center;
                yy = (int)(_1.y * scale) + bmp_center;
                //------------------------------------------------------------
                bmp_put_line_fast(&(p_space->bmp),
                                                       (int)(_0.x * scale) + bmp_center,
                                  p_space->bmp.yres - ((int)(_0.y * scale) + bmp_center),
                                  xx,
                                  p_space->bmp.yres - (yy),
                                  0x00000000
                                 );
                if(    p_space->show_vertices
                    && (    skin == 0
                         || p_space->show_onion_visuals
                       )
                  )
                    bmp_put_vertex(&(p_space->bmp),
                                   xx,
                                   p_space->bmp.yres - yy,
                                   0x00000000
                                  );
                //------------------------------------------------------------
                color = p_space->bmp.white;
                if(skin)
                {
                    fade = p_space->onion_skin_fade + (skin - 1) * (p_space->onion_skin_fade / 10);
                    if(fade > 1.0)
                        fade = 1.0;
                    color = color.blend(bmp_bg_color, fade);
                }
                if(    p_space->show_cursors
                    && is_index_selected(i)
                    && (    skin == 0
                         || p_space->show_onion_visuals
                       )
                  )
                    bmp_put_select(&(p_space->bmp),
                                   xx,
                                   p_space->bmp.yres - yy,
                                   color.as_uint()
                                  );
            }
        }
        //--------------------------------------------------------------------
        for(i = 1; i < quantity; i++)
        {
            if(at(i).is_color(p_space->black_level))
            {
                if(rotate_view)
                {
                    _0 = rotate_vertex(at(i - 1).bit_masked_position(p_space->signal_bit_mask), p_space->view_angle);
                    _1 = rotate_vertex(at(i    ).bit_masked_position(p_space->signal_bit_mask), p_space->view_angle);
                }
                else
                {
                    _0 = at(i - 1).bit_masked_position(p_space->signal_bit_mask);
                    _1 = at(i    ).bit_masked_position(p_space->signal_bit_mask);
                }
                if(offset_view)
                {
                    _0 += p_space->view_offset;
                    _1 += p_space->view_offset;
                }
                if(scale_view)
                {
                    _0 *= p_space->view_scale;
                    _1 *= p_space->view_scale;
                }
                xx = (int)(_1.x * scale) + bmp_center;
                yy = (int)(_1.y * scale) + bmp_center;
                color = at(i).bit_masked_color(p_space->signal_bit_mask);
                if(skin)
                {
                    fade = p_space->onion_skin_fade + (skin - 1) * (p_space->onion_skin_fade / 10);
                    if(fade > 1.0)
                        fade = 1.0;
                    color = color.blend(bmp_bg_color, fade);
                }
                //------------------------------------------------------------
                bmp_put_line_fast(&(p_space->bmp),
                                                       (int)(_0.x * scale) + bmp_center,
                                  p_space->bmp.yres - ((int)(_0.y * scale) + bmp_center),
                                  xx,
                                  p_space->bmp.yres - (yy),
                                  color.as_uint()
                                 );
                if(    p_space->show_vertices
                    && (    skin == 0
                         || p_space->show_onion_visuals
                       )
                  )
                    bmp_put_vertex(&(p_space->bmp),
                                   xx,
                                   p_space->bmp.yres - yy,
                                   color.as_uint()
                                  );
                //------------------------------------------------------------
                if(    p_space->show_cursors
                    && is_index_selected(i)
                    && (    skin == 0
                         || p_space->show_onion_visuals
                       )
                  )
                    bmp_put_select(   &(p_space->bmp)
                                    , xx
                                    , p_space->bmp.yres - yy
                                    , color.as_uint()
                                  );
            }
        }
        //--------------------------------------------------------------------
        // END ACTUAL VECTOR ART !!!
        //--------------------------------------------------------------------
    } // end if(quantity > 1)
    return;
}

//############################################################################
bool LaserBoy_frame::save_as_wave(string file)
{
    fstream out(file.c_str(), ios::out | ios::binary);
    //------------------------------------------------------------------------
    if(out.is_open())
    {
        if(quantity > 1)
        {
            int                   i,
                                  frame_scan,
                                  blank_dwell = 0,
                                  total_frame_scans;

            LaserBoy_vertex       vertex;

            LaserBoy_segment      intro  (p_space),
                                  bridge (p_space),
                                  coda   (p_space);

            LaserBoy_frame        copy(*this);

            LaserBoy_wave_header  header(p_space->sample_rate,
                                           LASERBOY_WAVE_POSITIVE
                                         | LASERBOY_WAVE_END_OF_FRAME
                                         | LASERBOY_WAVE_UNIQUE_FRAME
                                         | LASERBOY_WAVE_SIGNAL_MATRIX,
                                         p_space->channels_of_wav_out,
                                         LASERBOY_WAVE_VERSION
                                        );
            //----------------------------------------------------------------
            header.signal_id [0] = LASERBOY_SIGNAL_X_POSITION  ; // default values
            header.signal_id [1] = LASERBOY_SIGNAL_Y_POSITION  ;
            header.signal_id [2] = LASERBOY_SIGNAL_RED_ANALOG  ;
            header.signal_id [3] = LASERBOY_SIGNAL_GREEN_ANALOG;
            header.signal_id [4] = LASERBOY_SIGNAL_BLUE_ANALOG ;
            header.signal_id [5] = LASERBOY_SIGNAL_Z_POSITION  ;
            header.signal_id [6] = LASERBOY_SIGNAL_UNDEFINED   ;
            header.signal_id [7] = LASERBOY_SIGNAL_UNDEFINED   ;

            header.LSB_tag   [0] = LASERBOY_LSB_NOT_USED    ; // default values
            header.LSB_tag   [1] = LASERBOY_LSB_NOT_USED    ;
            header.LSB_tag   [2] = LASERBOY_LSB_END_OF_FRAME;
            header.LSB_tag   [3] = LASERBOY_LSB_UNIQUE_FRAME;
            header.LSB_tag   [4] = LASERBOY_LSB_NOT_USED    ;
            header.LSB_tag   [5] = LASERBOY_LSB_NOT_USED    ;
            header.LSB_tag   [6] = LASERBOY_LSB_NOT_USED    ;
            header.LSB_tag   [7] = LASERBOY_LSB_NOT_USED    ;

            if(    p_space->signal_bit_mask[0] != 0
                || p_space->signal_bit_mask[1] != 0
                || p_space->signal_bit_mask[2]  > 7
                || p_space->signal_bit_mask[3]  > 7
                || p_space->signal_bit_mask[4]  > 7
                || p_space->signal_bit_mask[5] != 0
                || p_space->signal_bit_mask[6] != 0
                || p_space->signal_bit_mask[7] != 0
              )
            {
                header.resolution[0] = 16 - p_space->signal_bit_mask[0];
                header.resolution[1] = 16 - p_space->signal_bit_mask[1];
                header.resolution[2] = 16 - p_space->signal_bit_mask[2];
                header.resolution[3] = 16 - p_space->signal_bit_mask[3];
                header.resolution[4] = 16 - p_space->signal_bit_mask[4];
                header.resolution[5] = 16 - p_space->signal_bit_mask[5];
                header.resolution[6] = 16 - p_space->signal_bit_mask[6];
                header.resolution[7] = 16 - p_space->signal_bit_mask[7];
                header.LaserBoy_wave_mode |= LASERBOY_SIGNAL_BIT_RESOLUTION;
            }
            //----------------------------------------------------------------
            if(p_space->view_angle != 0.0)
                copy.rotate_around_origin();
            //----------------------------------------------------------------
            if(copy.is_2D() || p_space->auto_flatten_z)
            {
                header.signal_id [5] = p_space->signal_if_not_z;
                header.resolution[5] = 9;
                copy.flatten_z();
            }
            //----------------------------------------------------------------
            if(p_space->auto_optimize)
            {
                header.LaserBoy_wave_mode |= LASERBOY_WAVE_OPTIMIZED;
                header.parms.lit_dwell_overhang     =        p_space->lit_dwell_overhang;
                header.parms.lit_delta_max          = (float)p_space->lit_delta_max;
                header.parms.blank_delta_max        = (float)p_space->blank_delta_max;
                header.parms.max_dwell_microsec     = (float)p_space->max_dwell_microsec;
                header.parms.insignificant_distance = (float)p_space->insignificant_distance;
                header.parms.insignificant_angle    = (float)p_space->insignificant_angle;
                header.parms.frames_per_second      = (float)p_space->frames_per_second;
                copy.optimize();
            }
            //----------------------------------------------------------------
            total_frame_scans = (int)ceil((   p_space->sample_rate
                                            * p_space->still_frame_duration
                                          ) / copy.quantity
                                         );
            //----------------------------------------------------------------
            if(p_space->auto_optimize)
            {
                blank_dwell = (int)(   p_space->sample_rate
                                     / p_space->max_dwell_microsec
                                   );
                //------------------------------------------------------------
                for(i = 0; i < blank_dwell; i++)
                    intro.push_back(LaserBoy_vertex()); // the origin
                //------------------------------------------------------------
                vertex = copy.first_lit_anchor();
                vertex.blank();
                intro += LaserBoy_segment(p_space, intro[0], vertex); // line from origin to first vertex
                //------------------------------------------------------------
                bridge = LaserBoy_segment(p_space, copy.last_lit_vertex(), vertex); // line from last vertex to first for repete
                bridge.blank_all_vertices();
                //------------------------------------------------------------
                coda = LaserBoy_segment(p_space, vertex, intro[0]); // line from last vertex to origin
                for(i = 0; i < blank_dwell; i++)
                    coda.push_back(LaserBoy_vertex()); // the origin
            } // end if(p_space->auto_optimize)
            //----------------------------------------------------------------
            if(p_space->invert_wave_output)
                header.LaserBoy_wave_mode &= ~LASERBOY_WAVE_POSITIVE;
            //----------------------------------------------------------------
            header.to_fstream_wave(out); // take up the space!
            //----------------------------------------------------------------
            if(p_space->auto_optimize)
                intro.to_fstream_wave(out, header, false, false);
            //----------------------------------------------------------------
            ((LaserBoy_segment)copy).to_fstream_wave(out, header, true, is_unique); // first frame unique
            //----------------------------------------------------------------
            for(frame_scan = 1; frame_scan < total_frame_scans; frame_scan++)
            {
                bridge.to_fstream_wave(out, header, false, false);
                ((LaserBoy_segment)copy).to_fstream_wave(out, header, true, false); // the rest are not
            }
            //----------------------------------------------------------------
            if(p_space->auto_optimize)
                coda.to_fstream_wave(out, header, false, false);
            //----------------------------------------------------------------
            header.to_fstream_wave(out);
            //----------------------------------------------------------------
        }
        out.close();
        return true;
    }
    return false;
}

//############################################################################
void LaserBoy_frame::to_fstream_wave(fstream& out, LaserBoy_wave_header& header)
{
    if(quantity > 1)
    {
        int frame_scan,
            total_frame_scans = 1;
        //--------------------------------------------------------------------
        LaserBoy_segment bridge(p_space, last_lit_vertex(), first_lit_anchor()); // line from last vertex to first for repete
        bridge.blank_all_vertices();
        //--------------------------------------------------------------------
        if(    p_space->auto_optimize
            && p_space->frames_per_second > 0.00
          )
        {
            total_frame_scans = (int)ceil(   (    p_space->sample_rate
                                                / (   quantity
                                                    + bridge.number_of_vertices()
                                                  )
                                              )
                                            / p_space->frames_per_second
                                         );
        } // end if(p_space->auto_optimize && p_space->frames_per_second > 0.00)
        //--------------------------------------------------------------------
        LaserBoy_segment::to_fstream_wave(out, header, true, true); // first pass through this frame
        //--------------------------------------------------------------------
        if(total_frame_scans > 1)
            for(frame_scan = 1; frame_scan < total_frame_scans; frame_scan++)
            {
                bridge.to_fstream_wave(out, header, false, false);
                LaserBoy_segment::to_fstream_wave(out, header, true, false);
            }
        //--------------------------------------------------------------------
        transition_to_next_frame.to_fstream_wave(out, header, false, false);
    }
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void LaserBoy_frame::to_ofstream_code(ofstream& out, const char* indent) const
{
    out << indent
        << "frame.format = "
        <<  (int)format
        <<  ";\n"
        <<  indent
        << "frame.palette_index = "
        <<  palette_index
        <<  ";\n";

    for(int i = 0; i < quantity; i++)
        out << indent
            <<  "frame.push_back(LaserBoy_vertex("
            << setw(6)
            <<  at(i).x
            <<  ", "
            << setw(6)
            <<  at(i).y
            <<  ", "
            << setw(6)
            <<  at(i).z
            <<  ", "
            << setw(3)
            <<  (int)at(i).r
            <<  ", "
            << setw(3)
            <<  (int)at(i).g
            <<  ", "
            << setw(3)
            <<  (int)at(i).b
            <<  ", "
            << setw(3)
            <<  (int)at(i).k
            <<  ", "
            << setw(3)
            <<  (int)at(i).c
            <<  "));\n";

    out << indent << "frame.quantity = (u_short)frame.quantity;\n";
    return;
}

//############################################################################
bool LaserBoy_frame::save_as_code(string object_name, string dir) const
{
    ofstream out((dir + object_name + ".cpp").c_str(), ios::out);
    if(out.is_open())
    {
        out << "//############################################################################\n"
               "// ~  This file is the product of LaserBoy!  ~\n"
               "//\n"
               "// by James Lehman\n"
               "// james@akrobiz.com\n"
               "//\n"
               "//############################################################################\n\n\n\n"
               "//############################################################################\n"

               "LaserBoy_frame "
            << object_name
            << "()\n"
               "{\n"
               "    LaserBoy_frame frame;\n";

        to_ofstream_code(out, "    ");

        out << "    return frame;\n"
               "}\n\n"
               "//############################################################################\n"
               "//////////////////////////////////////////////////////////////////////////////\n"
               "//############################################################################\n\n\n";

        out.close();
        return true;
    }
    return false;
}

//############################################################################
bool LaserBoy_frame::save_as_dxf(string file) const
{
    ofstream out(file.c_str(), ios::out);
    if(out.is_open())
    {
        int                 segment_vector_index,
                            vertex_index;
        LaserBoy_frame_set  segments = find_segments();
        //--------------------------------------------------------------------
        segments.minimize();
        //--------------------------------------------------------------------
        out << " 0\n"
                "SECTION\n"
                " 2\n"
                "ENTITIES\n";
        //--------------------------------------------------------------------
        for(segment_vector_index = 0; segment_vector_index < number_of_segments(); segment_vector_index++)
        {
            if(segments[segment_vector_index].number_of_color_vertices() == 1)
            {
                if(    (   segments[segment_vector_index].first_lit_anchor().position()
                         | segments[segment_vector_index].first_lit_vertex().position()
                       )
                    <= p_space->insignificant_distance
                  )
                {
                    out << " 0\n"
                           "POINT\n";
                    if(p_space->save_true_color_dxf)
                        out << " 420\n"
                            << segments[segment_vector_index].first_lit_vertex().color();
                    else
                        out << " 62\n"
                            << p_space->
                               palette_picker(LASERBOY_DXF)
                               .best_match(segments[segment_vector_index].first_lit_vertex());
                    out << "\n 10\n"
                        << segments[segment_vector_index].first_lit_anchor().x
                        << ".00\n 20\n"
                        << segments[segment_vector_index].first_lit_anchor().y
                        << ".00\n 30\n"
                        << segments[segment_vector_index].first_lit_anchor().z
                        << ".00\n";
                }
                else
                {
                    out << " 0\n"
                           "LINE\n";
                    if(p_space->save_true_color_dxf)
                        out << " 420\n"
                            << segments[segment_vector_index].first_lit_vertex().color();
                    else
                        out << " 62\n"
                            << p_space->
                               palette_picker(LASERBOY_DXF)
                               .best_match(segments[segment_vector_index].first_lit_vertex());
                    out << "\n 10\n"
                        << segments[segment_vector_index].first_lit_anchor().x
                        << ".00\n 20\n"
                        << segments[segment_vector_index].first_lit_anchor().y
                        << ".00\n 30\n"
                        << segments[segment_vector_index].first_lit_anchor().z
                        << ".00\n 11\n"
                        << segments[segment_vector_index].first_lit_vertex().x
                        << ".00\n 21\n"
                        << segments[segment_vector_index].first_lit_vertex().y
                        << ".00\n 31\n"
                        << segments[segment_vector_index].first_lit_vertex().z
                        << ".00\n";
                }
            }
            //----------------------------------------------------------------
            else
            {
                out << " 0\n"
                        "POLYLINE\n"
                        " 70\n"
                    << segments[segment_vector_index].is_closed_polygon()
                    << "\n 10\n"
                       "0.00\n"
                       " 20\n"
                       "0.00\n"
                       " 30\n"
                       "0.00\n"
                       //---------
                       " 0\n"
                       "VERTEX\n";
                if(p_space->save_true_color_dxf)
                    out << " 420\n"
                        << segments[segment_vector_index].at(1).color();
                else
                    out << " 62\n"
                        << p_space->
                           palette_picker(LASERBOY_DXF)
                           .best_match(segments[segment_vector_index].at(1));
                out << "\n 10\n"
                    << segments[segment_vector_index].first_lit_anchor().x
                    << ".00\n 20\n"
                    << segments[segment_vector_index].first_lit_anchor().y
                    << ".00\n 30\n"
                    << segments[segment_vector_index].first_lit_anchor().z
                    << ".00\n";

                for(vertex_index = 1; vertex_index < segments[segment_vector_index].number_of_vertices() - 1; vertex_index++)
                {
                    out << " 0\n"
                        << "VERTEX\n";
                    if(p_space->save_true_color_dxf)
                        out << " 420\n"
                            << segments[segment_vector_index].at(vertex_index + 1).color();
                    else
                        out << " 62\n"
                            << p_space->
                               palette_picker(LASERBOY_DXF)
                               .best_match(segments[segment_vector_index].at(vertex_index + 1));
                    out << "\n 10\n"
                        << segments[segment_vector_index].at(vertex_index).x
                        << ".00\n 20\n"
                        << segments[segment_vector_index].at(vertex_index).y
                        << ".00\n 30\n"
                        << segments[segment_vector_index].at(vertex_index).z
                        << ".00\n";
                }
                out << " 0\n"
                       "VERTEX\n";
                if(p_space->save_true_color_dxf)
                    out << " 420\n";
                else
                    out << " 62\n";
                out << "0\n"
                       " 10\n"
                    << segments[segment_vector_index].at(vertex_index).x
                    << ".00\n 20\n"
                    << segments[segment_vector_index].at(vertex_index).y
                    << ".00\n 30\n"
                    << segments[segment_vector_index].at(vertex_index).z
                    << ".00\n";
            }
        }
        //--------------------------------------------------------------------
        out << " 0\n"
               "ENDSEC\n"
               " 0\n"
               "EOF\n";
        //--------------------------------------------------------------------
        out.close();
        return true;
    }
    return false;
}

//############################################################################
bool LaserBoy_frame::save_as_bmp(string file) const
{
    if(number_of_vertices() > 1)
    {
        char   file_name[80];
        struct LaserBoy_bmp bmp_out = {0};

        if(palette_index == LASERBOY_TRUE_COLOR)
            bmp_init(&bmp_out, p_space->output_bmp_size, p_space->output_bmp_size, 24);
        else
        {
            bmp_init(&bmp_out, p_space->output_bmp_size, p_space->output_bmp_size,  8);
            p_space->palette_picker(palette_index).to_bmp_palette(&bmp_out);
        }
        bmp_clear(&bmp_out, bmp_out.black);

        u_short              i;
        LaserBoy_3D_double   _0,
                             _1;
        int                  bmp_center  = bmp_out.yres / 2;
        double               scale       = bmp_out.yres / 65536.0;
        //--------------------------------------------------------------------
        for(i = 1; i < quantity; i++)
        {
            if(at(i).is_color(p_space->black_level))
            {
                _0 = at(i - 1);
                _1 = at(i    );
                //------------------------------------------------------------

                if(palette_index == LASERBOY_TRUE_COLOR)
                    bmp_put_line_fast(   &bmp_out
                                       ,  (int)(_0.x * scale) + bmp_center
                                       , ((int)(_0.y * scale) + bmp_center)
                                       ,  (int)(_1.x * scale) + bmp_center
                                       , ((int)(_1.y * scale) + bmp_center)
                                       , bmp_out.make_rgb(at(i).r, at(i).g, at(i).b)
                                     );
                else
                    bmp_put_line_fast(   &bmp_out
                                       ,  (int)(_0.x * scale) + bmp_center
                                       , ((int)(_0.y * scale) + bmp_center)
                                       ,  (int)(_1.x * scale) + bmp_center
                                       , ((int)(_1.y * scale) + bmp_center)
                                       , at(i).c
                                     );
            }
        }
        sprintf(file_name, "%s", file.c_str());
        bmp_save_as(&bmp_out, file_name);
        bmp_free(&bmp_out);
        return true;
    }
    return false;
}

//############################################################################
LaserBoy_frame blank_frame(LaserBoy_space* ps)
{
    LaserBoy_frame frame(ps);
    frame = blank_segment(ps);
    frame.quantity = (u_short)frame.quantity;
    frame.is_2D();
    return frame;
}

//############################################################################
LaserBoy_frame NULL_frame(LaserBoy_space* ps)
{
    LaserBoy_frame frame(ps);
    frame = NULL_segment(ps);
    frame.quantity = (u_short)frame.quantity;
    frame.is_2D();
    return frame;
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
